#!/usr/bin/python3.7
import sys
import boto3
from boto3 import Session
from botocore.session import get_session
from botocore.credentials import DeferredRefreshableCredentials
import json
from datetime import datetime
import calendar
from asterisk.ami import AMIClient
from asterisk.ami import AutoReconnect
import time
import subprocess
import re
import traceback
import os
from ldap3 import Server, Connection, SAFE_SYNC
import secrets
import string
import syslog
import curses

re._pattern_type = re.Pattern

# TODO
# 1. Use environment variable for aws_access_key_id and aws_secret_access_key
# 2. add queue to push messages to kinesis, send it by batch to reduce payments
# 3. Add systemd service for use this script
# 4. Add all python's dependencies (boto3, asterisk.ami, ldap3) to deb package
# 5. Add cache for LDAP queries
# 6. Add BRIDGE_START logic for app queue (Local channels)

# CHECK
# 1. All kind of transfers
#   1.1. Hardware transfer (blind / attendant / semy attendant transfer) -- DONE
#   1.2. Collaboraation transfer (blind / attendant / semy attendant transfer) -- DONE
#   1.3. Feature сode (blind / attendant / semy attendant transfer) -- DONE
# 2. REFER after DirectRTP -- DONE
# 3. Pickup -- DONE
# 4. Forward -- DONE
# 5. Redirect -- DONE
######

# PBX
PBX_SERIAL = str(subprocess.getoutput("serial")).lower()
PBX_KEY = str(subprocess.getoutput("sed '18q;d' /rw2/etc/sf2"))
# Random password required to sign up in Cognito User Pool
PBX_RANDOM_PWD = ''.join((secrets.choice(string.ascii_letters + string.digits + string.punctuation) for i in range(16)))

# AWS
AWS_REGION = 'eu-central-1'
AWS_COGNITO_IDENTITY_POOL = AWS_REGION+':d593c2af-ecf7-4170-9919-1924edbe16bf'
AWS_COGNITO_IDENTITY_POOL_CLIENT = '575unmf4jm75sl0uum8u7hojio'
AWS_USER_POOL_ID = AWS_REGION+'_PpLQlcOZ6'
AWS_COGNITO_USER_POOL = 'cognito-idp.'+AWS_REGION+'.amazonaws.com/'+AWS_USER_POOL_ID
AWS_STREAM = 'wda-events-stage'

printEvent = True
STEP = 60

PrintSetX = dict()
PrintSetY = dict()

printFunction = print

class Screen:
    def __init__(self, start=False):
        self.status = start
        if start:
            self.screen = curses.initscr()
        else:
            self.screen = None

    def print(self, y, x, data):
        if self.screen is not None or not self.status:
            self.screen.addstr(y, x, data)

    def refresh(self):
        if self.screen is not None or not self.status:
            self.screen.refresh()

def print(arg):
    if screen.status:
        None
    else:
        printFunction(arg)

screen = Screen(printEvent)


# Auto-refreshable credential provider
class CredentialProvider():
    CANONICAL_NAME = "credential-provider"

    def load(self):
        #These creds will be automatically refreshed using the _refreh method if the current creds are going to expire
        creds = DeferredRefreshableCredentials(refresh_using=self._refresh, method="sts-assume-role",)
        return creds

    def METHOD():
        return "DeferredRefreshableCredentials"
    def _refresh(self):
        #Refresh AWS credtials
        response = self._aws_cred_refresh()
        credentials = {
            "access_key": response.get("AccessKeyId"),
            "secret_key": response.get("SecretKey"),
            "token": response.get("SessionToken"),
            "expiry_time": response.get("Expiration").isoformat(),
        }
        return credentials

    def _aws_cred_refresh(self):
        # Identity Provider Authorization (e.g. JWT for HTTP requests)
        identityProvider = boto3.client(
            'cognito-idp',
            region_name=AWS_REGION,
            aws_access_key_id='',
            aws_secret_access_key=''
        )
        try:
            # Initiate Identity Provider authentication flow
            authResponse = identityProvider.initiate_auth(
                AuthFlow='CUSTOM_AUTH',
                AuthParameters={
                    'USERNAME': PBX_SERIAL
                },
                ClientId=AWS_COGNITO_IDENTITY_POOL_CLIENT
            )
        except identityProvider.exceptions.UserNotFoundException:
            # Register the PBX in the Cognito User Pool
            identityProvider.sign_up(
                ClientId=AWS_COGNITO_IDENTITY_POOL_CLIENT,
                Password=PBX_RANDOM_PWD,
                Username=PBX_SERIAL,
                ValidationData=[
                    {
                        'Name': 'key',
                        'Value': PBX_KEY,
                    },
                ],
            )

            # Initiate Identity Provider authentication flow
            authResponse = identityProvider.initiate_auth(
                AuthFlow='CUSTOM_AUTH',
                AuthParameters={
                    'USERNAME': PBX_SERIAL
                },
                ClientId=AWS_COGNITO_IDENTITY_POOL_CLIENT
            )

        # Get auth result - IdToken, AccessToken etc.
        authResult = identityProvider.respond_to_auth_challenge(
            ChallengeName='CUSTOM_CHALLENGE',
            ChallengeResponses={
                'USERNAME': PBX_SERIAL,
                'ANSWER': PBX_KEY
            },
            ClientId=AWS_COGNITO_IDENTITY_POOL_CLIENT,
            Session=authResponse['Session']
        )

        if 'AuthenticationResult' not in authResult:
            print('Authorization failed: {}'.format(authResult))
            raise Exception("Cognito authentication failed")

        # print("Cognito Credentials:\n{}\n\n\n\n".format(authResult['AuthenticationResult']))

        # Identity Authorization (e.g. Token to access AWS services from PBX role like S3, Kinesis)
        clientIdentity = boto3.client('cognito-identity', region_name=AWS_REGION)

        # Get Cognito Identity ID
        identityResponse = clientIdentity.get_id(
          IdentityPoolId=AWS_COGNITO_IDENTITY_POOL,
          Logins={
            AWS_COGNITO_USER_POOL: authResult['AuthenticationResult']['IdToken']
          }
        )
        # Get credentials for the Identity
        credentialsResponse = clientIdentity.get_credentials_for_identity(
          IdentityId=identityResponse['IdentityId'],
          Logins={
            AWS_COGNITO_USER_POOL: authResult['AuthenticationResult']['IdToken']
          }
        )

        # print("AWS Credentials:\n{}\n\n\n\n".format(credentialsResponse["Credentials"]))
        return credentialsResponse["Credentials"]

# AWS Credential Provider object
AWS_CREDENTIAL_PROVIDER = CredentialProvider()

# AMIClient
API_CLIENT = AMIClient(address='127.0.0.1', port=5039)

# LDAP
LDAP_USER = subprocess.getoutput("cat /etc/callweaver/res_ldap.conf | grep 'user=' | cut -d= -f 2,3,4")
LDAP_PASS = subprocess.getoutput("cat /etc/callweaver/res_ldap.conf | grep 'pass=' | cut -d= -f 2")

LDAP_SERVER = Server("localhost")

LDAP_CONN = Connection(LDAP_SERVER, LDAP_USER, LDAP_PASS, client_strategy=SAFE_SYNC, auto_bind=True)


#####


# AMI functions
def on_disconnect(client, error):
    if error is not None:
        print("Connection error: " + str(error))


AUTO_RECONNECT = AutoReconnect(ami_client=API_CLIENT, delay=20, on_disconnect=on_disconnect)

ChannelConversation = dict()
ListConversations = dict()
gListChannels = dict()
reinit_channels = set()


class ChannelOwnerType:
    USER = 1
    CUSTOMER = 2


class ChannelType:
    SIP = 1
    LOCAL = 2


class Channel:
    class InternalData:
        def __init__(self, ownerType, ownerID, ownerEmail, ownerExt, ownerGroupID, ownerGroupName, chanType, name, ID):
            self.ownerType = ownerType
            self.ownerID = ownerID
            self.ownerEmail = ownerEmail
            self.ownerExt = ownerExt
            self.ownerGroupID = ownerGroupID
            self.ownerGroupName = ownerGroupName
            self.type = chanType
            self.name = name
            self.publicID = ID
            self.flagSkipPublish = False

        def __str__(self):
            return "ownerType: {}, ownerID: {}, ownerEmail: {}, ownerExt: {}, ownerGroupID: {}, ownerGroupName: {}, type: {}, name: {}, publicID: {}".format(
                self.ownerType, self.ownerID, self.ownerEmail, self.ownerExt, self.ownerGroupID, self.ownerGroupName,
                ("USER" if self.type == ChannelOwnerType.USER else "CUSTOMER"), self.name, self.publicID)

    def __init__(self, ID, ownerType, ownerID, ownerEmail, ownerExt, ownerGroupID, ownerGroupName, chanType, name):
        self.ID = ID
        self._private_data = self.InternalData(ownerType=ownerType,
                                               ownerID=ownerID,
                                               ownerEmail=ownerEmail,
                                               ownerExt=ownerExt,
                                               ownerGroupID=ownerGroupID,
                                               ownerGroupName=ownerGroupName,
                                               chanType=chanType,
                                               name=name,
                                               ID=ID)
        gListChannels[ID] = self

    def __del__(self):
        # print("====Destructor!!! Delete Channel {} name: {}".format(self.ID, self.getName()))
        return None

    def getName(self):
        return self._private_data.name

    def getOwnerType(self):
        return self._private_data.ownerType

    def getOwnerID(self):
        return self._private_data.ownerID

    def getOwnerEmail(self):
        return self._private_data.ownerEmail

    def getOwnerExt(self):
        return self._private_data.ownerExt

    def getOwnerGroupID(self):
        return self._private_data.ownerGroupID

    def getOwnerGroupName(self):
        return self._private_data.ownerGroupName

    def getType(self):
        return self._private_data.type

    def setFlagSkipPublish(self, value):
        self._private_data.flagSkipPublish = value

    def getFlagSkipPublish(self):
        return self._private_data.flagSkipPublish

    def getPublicID(self):
        return self._private_data.publicID

    def switchData(self, chan):
        tmp = self._private_data
        # self._private_data = chan._private_data
        chan._private_data = tmp
        # print("======After: \nChan1: ({}) \nChan2: ({})".format(self, chan))

    def __str__(self):
        return "ID: {}, Data: {}".format(self.ID, str(self._private_data))


class Conversation:
    def __init__(self, ID):
        # print("====Init conversation " + str(ID))
        self.ID = ID
        self.listChannels = dict()
        self.remove = False
        ListConversations[ID] = self

    def __str__(self):
        res = "ID: {}, remove: {}".format(self.ID, self.remove)
        for c in self.listChannels.keys():
            res += "\n\t\t{}: {}".format(c, self.listChannels[c])
        return res

    def __del__(self):
        # print("====Destructor!!! Delete conversation {}".format(self))
        # for ID in self.listChannels.keys():
        #     self.unlinkChannel(ID)
        None

    def linkChannel(self, chan: Channel):
        # print("====Link channel {} ({}) to conversation: {}".format(chan.ID, chan.getName(), self.ID))
        self.listChannels[chan.ID] = chan
        self.remove = False
        gListChannels[chan.ID] = chan
        ChannelConversation[chan.ID] = self.ID

    def relinkChannel(self, chan: Channel):
        if ChannelConversation[chan.ID] == self.ID:
            return
        conv = getConversation(chan.ID)
        if conv is None:
            return
        print("relink {}".format(chan))
        conv.unlinkChannel(chan.ID)
        self.linkChannel(chan)

    def unlinkChannel(self, chanID):
        chan = getChannel(chanID)
        if not chan:
            print("ERROR: Channel: {} not found".format(chanID))
            return
        # print("====Unlink channel {} ({}) from conversation: {}".format(chan.ID, chan.getName(), self.ID))
        print_currect_state()
        if chanID in self.listChannels.keys():
            del self.listChannels[chanID]

        if chanID in ChannelConversation.keys():
            del ChannelConversation[chanID]

        if chanID in gListChannels.keys():
            del gListChannels[chanID]

        if len(self.listChannels.keys()) == 0 and self.ID in ListConversations.keys():
            if ListConversations[self.ID].remove:
                del ListConversations[self.ID]
            else:
                ListConversations[self.ID].remove = True


def getConversation(chan):
    # print("====Try to find conversation: {}".format(chan))
    if chan in ListConversations.keys():
        return ListConversations[chan]
    if not chan in ChannelConversation.keys():
        print("ERROR: Conversation for {} not found (no channel in ChannelConversation)".format(chan))
        print_currect_state()
        return None
    elif not ChannelConversation[chan] in ListConversations.keys():
        print("ERROR: Conversation for channel {} not found (conversation {} doesn't exist in ListConversations)".format(chan, ChannelConversation[chan]))
        print_currect_state()
        return None
    return ListConversations[ChannelConversation[chan]]


def getChannel(chan):
    if chan in gListChannels.keys():
        return gListChannels[chan]
    else:
        # print("WARNING: Channel {} not found".format(chan))
        print_currect_state()
        return None


def getUserInfo(number):
    rv = {}
    status, _, response, _ = LDAP_CONN.search(search_base="dc=wildix",
                                              search_filter="(&(objectClass=wPerson)(|(telephoneNumber={})(mobile={})))".format(
                                                  number, number),
                                              attributes=["uid", "mail", "telephoneNumber", "owner", "userKind"])
    if status is True and len(response) > 0:
        rv["type"] = ChannelOwnerType.USER
        if len(response[0].get("attributes").get("uid")) > 0:
            rv["ID"] = response[0].get("attributes").get("uid")[0]
        if len(response[0].get("attributes").get("mail")) > 0:
            rv["email"] = response[0].get("attributes").get("mail")[0]
        if len(response[0].get("attributes").get("telephoneNumber")) > 0:
            rv["ext"] = response[0].get("attributes").get("telephoneNumber")[0]
        if len(response[0].get("attributes").get("uid")) > 0:
            gid = re.findall("^gid=(\d+),.*", response[0].get("attributes").get("owner")[0])
            rv["groupID"] = gid[0] if len(gid) > 0 else None
        if len(response[0].get("attributes").get("userKind")) >0:
            rv["userKind"] = response[0].get("attributes").get("userKind")[0]
    return rv

def getGroupInfo(gid):
    rv = {}
    status, _, response, _ = LDAP_CONN.search(search_base="dc=wildix",
                                              search_filter="(&(objectClass=wGroup)(gid={}))".format(gid),
                                              attributes=["cn"])
    if status is True and len(response) > 0:
        rv["id"] = gid
        if len(response[0].get("attributes").get("cn")) > 0:
            rv["name"] = response[0].get("attributes").get("cn")[0]
    return rv

def getCustomerInfo(number):
    rv = {}
    if number is not None and number != "":
        rv["type"] = ChannelOwnerType.CUSTOMER
    return rv


def handler_event_masquerade(event):
    c1 = get_param(event, "UniqueID1")
    c2 = get_param(event, "UniqueID2")
    chan1 = getChannel(c1)
    chan2 = getChannel(c2)
    if chan1 is None or chan2 is None:
        print("ERROR: Channel {} NOT FOUND".format(c1 if chan1 is None else c2))
        return
    chan1.switchData(chan2)
    conversation = getConversation(get_param(event, "LinkedID"))
    conversation.relinkChannel(chan1)
    conversation.relinkChannel(chan2)


def handler_cel(event):
    # CHAN_START
    if get_param(event, "EventName") == "CHAN_START":
        UniqueID = get_param(event, "UniqueID")
        LinkedID = get_param(event, "LinkedID")
        CallerIDnum = get_param(event, "CallerIDnum")
        CallerIDname = get_param(event, "CallerIDname")
        ownerInfo = getUserInfo(CallerIDnum)
        if len(ownerInfo) == 0:
            ownerInfo = getCustomerInfo(CallerIDnum)
        if UniqueID == LinkedID:
            conversation = Conversation(UniqueID)
        else:
            conversation = getConversation(LinkedID)
        if "groupID" in ownerInfo.keys():
            groupInfo = getGroupInfo(ownerInfo["groupID"])
            ownerInfo["groupName"] = groupInfo.get("name")
        channel = get_param(event, "Channel")
        chan = Channel(ID=UniqueID,
                       ownerType=ownerInfo.get("type"),
                       ownerID=ownerInfo.get("ID"),
                       ownerEmail=ownerInfo.get("email"),
                       ownerExt=ownerInfo.get("ext"),
                       ownerGroupID=ownerInfo.get("groupID"),
                       ownerGroupName=ownerInfo.get("groupName"),
                       chanType=ChannelType.SIP if channel.startswith("SIP/") else ChannelType.LOCAL,
                       name=channel)
        conversation.linkChannel(chan)
        if channel in reinit_channels:
            reinit_channels.remove(channel)
            return
        if UniqueID != LinkedID:
            if chan.getType() == ChannelType.LOCAL:
                chan.setFlagSkipPublish(True)
                return

        push_event(EventItem(
            type="CALL.CHANNEL",
            uid=UniqueID,
            lid=LinkedID,
            data = {
                "userId":None,
                "userEmail":None,
                "userExtension":None,
                "groupId":None,
                "groupName":None,
                "phone":CallerIDnum,
                "name":CallerIDname,
                "destination":get_param(event, "Exten")
            }
        ))
    # TRUNK_CALL
    elif get_param(event, "EventName") == "TrunkCall":
        extra = get_param(event, "Extra").split(",")
        push_event(EventItem(
            type="CALL.TRUNK",
            uid=get_param(event, "UniqueID"),
            lid=get_param(event, "LinkedID"),
            data={
                "direction":extra[0],
                "trunkName":extra[1],
            }
        ))
    # CHAN_END
    elif get_param(event, "EventName") == "CHAN_END":
        UniqueID = get_param(event, "UniqueID")
        LinkedID = get_param(event, "LinkedID")
        conversation = getConversation(UniqueID)
        if conversation:
            conversation.unlinkChannel(UniqueID)
    # HANGUP
    elif get_param(event, "EventName") == "HANGUP":
        UniqueID = get_param(event, "UniqueID")
        chan = getChannel(UniqueID)
        if not chan or chan.getFlagSkipPublish():
            # print("====Skip sending event HANGUP for {}".format(UniqueID))
            return
        # avoid sending duplicating HANGUP event
        if chan.getPublicID() != UniqueID:
            return
        c = get_param(event, "Channel")
        extra = get_param(event, "Extra")
        splitExtra = extra.split(",")
        hangupcause = splitExtra[0]
        hangupsource = splitExtra[1]
        dialstatus = splitExtra[2]
        causeStr = splitExtra[3]
        if len(splitExtra) > 4:
            for i in range(4, len(splitExtra)):
                causeStr += splitExtra[i]
        source = "remote"
        if hangupsource == c:
            source = "local"

        push_event(EventItem(
            type="CALL.HANGUP",
            uid=get_param(event, "UniqueID"),
            lid=get_param(event, "LinkedID"),
            data={
                "dialStatus":dialstatus if dialstatus != "" else None,
                "source":source,
                "cause":int(hangupcause),
                "causeStr":causeStr if causeStr != "Unknown" else None
            }
        ))
    # LOOKUP
    elif get_param(event, "EventName") == "LookUp":
        extra = get_param(event, "Extra").split(",")
        push_event(EventItem(
            type="CALL.LOOKUP",
            uid=get_param(event, "UniqueID"),
            lid=get_param(event, "LinkedID"),
            data={
                "name":extra[0],
                "company":extra[1] if extra[1] != "" else None,
                "phone":extra[2] if (len(extra)>2 and extra[2] != "") else None,
                "customerEmail":extra[3] if (len(extra)>3 and extra[3] != "") else None,
            }
        ))
    elif get_param(event, "EventName") == "ANSWER":
        UniqueID = get_param(event, "UniqueID")
        chan = getChannel(UniqueID)
        if chan.getFlagSkipPublish():
            return
        push_event(EventItem(
            type="CALL.ANSWER",
            uid=UniqueID,
            lid=get_param(event, "LinkedID")))
    # ADD_TAG
    elif get_param(event, "EventName") == "Tag":
        push_tag_event("CALL.TAG", event)
    # REMOVE_TAG
    # elif get_param(event, "EventName") == "RemoveTag":
    #     push_tag_event("REMOVE_TAG", event)
    # LINKEDID_END
    elif get_param(event, "EventName") == "LINKEDID_END":
        UniqueID = get_param(event, "UniqueID")
        LinkedID = get_param(event, "LinkedID")
        for i in (UniqueID, LinkedID):
            if i in ListConversations.keys():
                if len(ListConversations[i].listChannels.keys()) == 0 and ListConversations[i].remove:
                    del ListConversations[i]
                else:
                    ListConversations[i].remove = True
        push_event(EventItem(
            type="CALL.LINKEDID_END",
            uid=get_param(event, "UniqueID"),
            lid=get_param(event, "LinkedID"),
            data={}
        ))
    # CALL_GROUP
    elif get_param(event, "EventName") == "CallGroup":
        extra = get_param(event, "Extra").replace(" ","").split(",")
        push_event(EventItem(
            type="CALL.CALL_GROUP",
            uid=get_param(event, "UniqueID"),
            lid=get_param(event, "LinkedID"),
            data={
                "name":extra[0] if extra[0] != "" else None,
                "id":extra[1] if (len(extra)>1 and extra[1] != "") else None
            }
        ))
    # PICKUP | PARK_END
    elif get_param(event, "EventName") == "PICKUP":
        UniqueID = get_param(event, "UniqueID")
        LinkedID = get_param(event, "LinkedID")
        destination = get_param(event, "Exten")
        destInfo = getUserInfo(destination)
        # PARK_END
        if "userKind" in destInfo.keys() and "park_orbit" == destInfo["userKind"]:
            push_park_event({"action":"end","destination":destination}, event)
        # PICKUP
        else:
            push_event(EventItem(
                type="CALL.PICKUP",
                uid=UniqueID,
                lid=LinkedID,
                data={"destination":destination}
            ))
    # FORWARD
    elif get_param(event, "EventName") == "FORWARD":
        UniqueID = get_param(event, "UniqueID")
        LinkedID = get_param(event, "LinkedID")
        extra = get_param(event, "Extra").split(",")
        push_event(EventItem(
            type="CALL.FORWARD",
            uid=UniqueID,
            lid=LinkedID,
            data={"destination":extra[0]}
        ))
    # TRANSFER | PARK_START
    elif get_param(event, "EventName") in {"BLINDTRANSFER", "ATTENDEDTRANSFER"}:
        UniqueID = get_param(event, "UniqueID")
        LinkedID = get_param(event, "LinkedID")
        transferType = re.findall("^(\w+)TRANSFER", get_param(event, "EventName"))[0].lower()
        extra = get_param(event, "Extra").split(",")
        if transferType == "blind":
            destination = extra[0]
        elif transferType == "attended":
            destination = get_param(event, "Exten")
        destInfo = getUserInfo(destination)
        # PARK_START
        if "userKind" in destInfo.keys() and "park_orbit" == destInfo["userKind"]:
            push_park_event({"action":"start","destination":destination}, event)
        # TRANSFER
        else:
            push_event(EventItem(
                type="CALL.TRANSFER",
                uid=UniqueID,
                lid=LinkedID,
                data={
                    "destination":destination,
                    "transferType":transferType
                }
            ))
    # else:
    #     print(event)


def event_listener(event, **kwargs):
    # print("====----------- Got event: {}".format(event))
    try:
        if event.name == "CEL":
            handler_cel(event)
        elif event.name == "MusicOnHold":
            push_event(EventItem(
                type="CALL.MOH",
                uid=get_param(event, "UniqueID"),
                lid=get_param(event, "LinkedID"),
                data={
                    "state":get_param(event, "State")
                }
            ))
        # RECORD_START
        elif event.name == "MonitorStart":
            UniqueID = get_param(event, "UniqueID")
            LinkedID = get_param(event, "LinkedID")
            push_event(EventItem(
                type="CALL.RECORD_START",
                uid=UniqueID,
                lid=LinkedID)
            )
        # RECORD_STOP
        elif event.name == "MonitorStop":
            UniqueID = get_param(event, "UniqueID")
            LinkedID = get_param(event, "LinkedID")
            push_event(EventItem(
                type="CALL.RECORD_STOP",
                uid=UniqueID,
                lid=LinkedID,
                data={"fileName":get_param(event, "FileName")}
            ))
        elif event.name == "UniqueIDMasquerade":
            handler_event_masquerade(event)
        elif event.name == "REINIT_CHANNEL":
            reinit_channels.add(get_param(event, "Channel"))
        # BRIDGE
        elif event.name in {"BRIDGE_START", "BRIDGE_UPDATE", "BRIDGE_END"}:
            UniqueID = get_param(event, "UniqueID1")
            UniqueID2 = get_param(event, "UniqueID2")
            chan = getChannel(UniqueID)
            if chan.getFlagSkipPublish():
                return
            chan = getChannel(UniqueID2)
            if chan.getFlagSkipPublish():
                return
            action = re.findall("^BRIDGE_(\w+)", event.name)[0].lower()
            push_event(EventItem(
                type="CALL.BRIDGE",
                uid=UniqueID,
                lid=get_param(event, "LinkedID"),
                data={
                    "action":action,
                    "with":UniqueID2
                }
            ))
        # else:
        #     print(event)
    except AttributeError as e1:
        print("ERROR: {}".format(e1))
        print_currect_state()
    except Exception as e:
        print("ERROR: ", e, str(traceback.format_exc()))

def init_aws_client():
    check_cognito_credentials()
    session = get_session()
    cred_provider = session.get_component('credential_provider')
    cred_provider.insert_before('env', AWS_CREDENTIAL_PROVIDER)
    session.set_config_variable("region", AWS_REGION)
    autorefresh_session = Session(botocore_session=session)

    # Kinesis client
    global KINESIS_CLIENT
    KINESIS_CLIENT = autorefresh_session.client('kinesis', region_name=AWS_REGION)

def check_cognito_credentials():
    try:
        AWS_CREDENTIAL_PROVIDER._refresh()
    except Exception as e:
        print(e)
        syslog.syslog(syslog.LOG_ERR, '[call-statistic][auth_challenge] Cognito authentication failed.')
        stop(3600)

def init_connection():
    API_CLIENT.login(username='root', secret='wildix')
    # API_CLIENT.add_event_listener(event_listener, white_list=[
    #     "CEL",
    #     "MonitorStart",
    #     "MonitorStop",
    #     "UniqueIDMasquerade",
    #       "MusicOnHold",
    #       "BRIDGE_START",
    #       "BRIDGE_STOP",
    #       "REINIT_CHANNEL"
    # ])
    API_CLIENT.add_event_listener(event_listener, black_list=[
        "DeviceState",
        "Newchannel",
        "ChannelUpdate",
        "Newstate",
        "EventConnectedLine",
        "Bridge",
        "NewCallerid",
        "ExtensionStatus",
        "NewAccountCode",
        "HangupRequest",
        "SoftHangupRequest",
        "PeerStatus",
        "Cdr",
        "Hold",
        "QueueMemberStatus",
        "Dial",
        "Masquerade",
        "Rename",
        "BridgeAction",
        "Hangup",
        "EventFeature"  # {'Privilege': 'call,all', 'Device': '102', 'Feature': 'AWAY', 'State': 'false'}
    ])

def stop(delay = None):
    msg = ""
    if delay:
        msg = " with a delay of "+str(delay)+" seconds"
    print("Stop{}".format(msg))

    API_CLIENT.logoff()

    if printEvent:
        curses.endwin()

    if delay:
        time.sleep(delay)
    sys.exit()

###########


def getPublicID(UniqueID):
    chan = getChannel(UniqueID)
    if chan is not None:
        return chan.getPublicID()
    return UniqueID


# BOTO3
class EventItem(dict):
    def __init__(self, uid, **kwargs):
        dict.__init__(self,
                      time=calendar.timegm(datetime.utcnow().timetuple()),
                      pbx=PBX_SERIAL)
        chan = getChannel(uid)
        data = kwargs.get("data",{})
        if chan is not None:
            if chan.getOwnerType() == ChannelOwnerType.USER and len(data) > 0:
                if "userId" in data.keys():
                    data["userId"] = chan.getOwnerID()
                if "userEmail" in data.keys():
                    data["userEmail"] = chan.getOwnerEmail()
                if "userExtension" in data.keys():
                    data["userExtension"] = chan.getOwnerExt()
                if "groupId" in data.keys():
                    data["groupId"] = chan.getOwnerGroupID()
                if "groupName" in data.keys():
                    data["groupName"] = chan.getOwnerGroupName()
            elif chan.getOwnerType() == ChannelOwnerType.CUSTOMER and len(data) > 0:
                if "customerEmail" in data.keys():
                    data["customerEmail"] = data["customerEmail"] if data["customerEmail"] != None else chan.getOwnerEmail()
            self["uid"] = normalize("uid", chan.getPublicID())
        else:
            if kwargs.get("type") != "CALL.LINKEDID_END":
                print("ERROR: !!!!! Channel {} NOT FOUND".format(uid))
            print_currect_state()
            self["uid"] = normalize("uid", uid)
        for key, value in kwargs.items():
            if key.startswith("uid"):
                self[key] = normalize(key, getPublicID(value))
            elif key.startswith("data"):
                self["data"] = {}
                for k, v in data.items():
                    if v is not None and v != "":
                        self["data"][k] = normalize(k, v)
            else:
                if value is not None and value != "":
                    self[key] = normalize(key, value)
        print_screen(time, self.get("type"), self["uid"], self["lid"], data)


def get_param(event, name):
    if name in event.keys:
        return event[name]
    elif name == "UniqueID":
        return get_param(event, "Uniqueid")
    elif name == "LinkedID":
        return get_param(event, "Linkedid")
    return None

def normalize(key, value):
    rv = value
    if key in {"uid","lid","with"} and type(value) is str:
        rv = value.replace("wildixbox-","")
    return rv

def push_tag_event(name, event):
    push_event(EventItem(
        type=name,
        uid=get_param(event, "UniqueID"),
        lid=get_param(event, "LinkedID"),
        data={"tags":get_param(event, "Extra").replace(" ", "").split(",") if get_param(event, "Extra") is not None else None}
    ))

def push_park_event(data, event):
    push_event(EventItem(
        type="CALL.PARK",
        uid=get_param(event, "UniqueID"),
        lid=get_param(event, "LinkedID"),
        data=data
    ))

def print_screen(time, type, uid, lid, json_data):
    if not screen.status:
        return
    if lid not in PrintSetX:
        PrintSetX[lid] = STEP + (2 * len(PrintSetX) * STEP)
        screen.print(0, PrintSetX[lid] - 3 - int(len(lid) / 2), lid)
        PrintSetY[lid] = 1
    x = PrintSetX.get(lid)
    y = PrintSetY.get(lid)
    off = STEP - 3
    left_format = "{0:{1}.{2}s} |"
    right_format = "| {0:{1}.{2}s}"
    if lid == uid:
        x = x - STEP
        data1 = left_format.format(type, off, off)
        data2 = None
        if len(json_data) > 0:
            data2 = left_format.format(str(json_data), off, off)
    else:
        x = x - 2
        data1 = right_format.format(str(type) + " (" + uid + ")", off , off)
        data2 = None
        if len(json_data) > 0:
            data2 = right_format.format(str(str(json_data)), off, off)
    screen.print(y, x, data1)
    if data2 is not None:
        screen.print(y + 1, x, data2)
    PrintSetY[lid] = y + (2 if data2 is not None else 1)
    screen.refresh()
    # print("time: {}, type: {}, uid: {}, lid: {}, data: {}".format(time, type, uid, lid, data))

def push_event(data):
    payload = json.dumps(data)
    print(payload)
    payload_bytes = payload.encode('utf-8')
    put_response = KINESIS_CLIENT.put_record(
        StreamName=AWS_STREAM,
        Data=payload_bytes,
        PartitionKey=PBX_SERIAL)

    # print("Response: " + str(put_response))


###########


def print_currect_state():
    # print("====ChannelConversation: {}".format(ChannelConversation))
    res = "ListConversations:"
    for c in ListConversations.keys():
        res += "\t{}".format(ListConversations[c])
    # print("===={}".format(res))
    res = "gListChannels:"
    for c in gListChannels.keys():
        res += "\t{}".format(gListChannels[c])
    # print("===={}".format(res))
    # print("====reinit_channels: {}".format(reinit_channels))
    return None

# def garbageCollector():
#     for c in ListConversations.keys():
#         if c.remove == True and c.listChannels.keys() == 0

def main():
    print("Start")

    init_connection()

    init_aws_client()

    try:
        while True:
            print_currect_state()
            check_cognito_credentials()
            time.sleep(20)
    except (KeyboardInterrupt, SystemExit):
        None

    stop()

if __name__ == "__main__":
    sys.exit(main())

